@@ -14,23 +14,22 @@ use MooseX::Params::Validate ();
 
 use Params::Registry::Template;
 use Params::Registry::Instance;
+use Params::Registry::Error;
 
 use URI;
 use URI::QueryParam;
 
-with 'Throwable';
-
 =head1 NAME
 
 Params::Registry - Housekeeping for sets of named parameters
 
 =head1 VERSION
 
-Version 0.01
+Version 0.03
 
 =cut
 
-our $VERSION = '0.01';
+our $VERSION = '0.03';
 
 =head1 SYNOPSIS
 
@@ -174,14 +173,17 @@ around BUILDARGS => sub {
             $map{$name} = $use;
         }
         # TODO: throw a proper error on duplicate key
-        die if exists $p{params}{$name};
+        Params::Registry::Error->throw
+              ("Parameter $name already exists") if exists $p{params}{$name};
+
         $p{params}{$name} = $entry;
     }
 
     # second pass to stitch the reused parameters together
     while (my ($k, $v) = each %map) {
         # TODO throw a proper error if the target isn't found
-        my $p = $p{params}{$v} or die;
+        my $p = $p{params}{$v} or Params::Registry::Error->throw
+            ("Tried to resolve $v for reuse but couldn't find it");
 
         # overwrite with any new data
         $p{params}{$k} = {%$p, %{$p{params}{$k}}};
@@ -203,7 +205,9 @@ sub BUILD {
         my %t = %{$p->{$k}};
         my $x = $p->{$k} = Params::Registry::Template->new
             (%t, registry => $self);
-        if ($x->consumes > 0) {
+        if ($x->_consdep > 0) {
+            # shortcut because only parameters with dependencies will
+            # have a rank higher than zero
             $rank{$k} = 1;
             push @stack, $k;
         }
@@ -214,15 +218,15 @@ sub BUILD {
 
     # construct a rank tree
 
-    my %seen;
+    #my %seen;
     while (my $x = shift @stack) {
         #my %c = map { $_ => 1 } $p->{$x}->consumes;
         my $match = 0;
-        for my $c ($p->{$x}->consumes) {
+        for my $c ($p->{$x}->_consdep) {
             $match = 1 if $rank{$x} == $rank{$c};
             # XXX will this actually catch all cycles?
-            Carp::croak("Cycle detected between $x and $c")
-                  if $rank{$x} < $rank{$c};
+            Params::Registry::Error->throw
+                  ("Cycle detected between $x and $c") if $rank{$x} < $rank{$c};
         }
 
         if ($match) {
@@ -347,7 +351,7 @@ sub process {
             $obj = $_[0];
         }
         else {
-            $self->throw
+            Params::Registry::Error->throw
                 ('If the argument is a ref, it must be a URI or a HASH ref');
         }
     }
@@ -361,7 +365,8 @@ sub process {
         $obj = \%x;
     }
     else {
-        $self->throw('Check your inputs to Params::Registry::process');
+        Params::Registry::Error->throw
+              ('Check your inputs to Params::Registry::process');
     }
 
     my $instance = Params::Registry::Instance->new(registry => $self);
@@ -431,11 +436,24 @@ L<http://search.cpan.org/dist/Params-Registry/>
 
 =back
 
-
 =head1 SEE ALSO
 
+=over 4
+
+=item
+
+L<Params::Registry::Instance>
+
+=item
+
+L<Params::Registry::Template>
+
+=item
+
 L<Params::Validate>
 
+=back
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2013 Dorian Taylor.
